"""
Unit tests for input adapters.
"""

import json
import socket
import time
from typing import Callable, Dict
from unittest.mock import MagicMock, Mock, patch

from multilat_solver import INPUT_ADAPTERS, InputAdapter
from multilat_solver.input_adapters import (
    CircularBuffer,
    FileInputAdapter,
    Message,
    SerialInputAdapter,
    UDPInputAdapter,
    register_input_adapter,
)


class TestMessage:
    """Test Message parser."""

    def test_message_parsing(self):
        """Test message parsing from bytes."""
        # Create a test message: id=1, data=5000 (in mm)
        data = b"\x01\x88\x13\x00\x00"  # Little-endian: id=1, int=5000
        msg = Message(data)
        assert msg.id == 1
        assert msg.data == 5000


class TestCircularBuffer:
    """Test CircularBuffer."""

    def test_buffer_append_pop(self):
        """Test basic append and pop."""
        buffer = CircularBuffer(10, element_size=5)
        assert buffer.size == 0

        buffer.append(b"test")
        assert buffer.size == 1

        item = buffer.pop()
        assert item == b"test"
        assert buffer.size == 0

    def test_buffer_wrap_around(self):
        """Test buffer wraps around correctly."""
        buffer = CircularBuffer(3, element_size=5)

        buffer.append(b"1")
        buffer.append(b"2")
        buffer.append(b"3")
        assert buffer.size == 3

        buffer.append(b"4")  # Should wrap around
        assert buffer.size == 3

        assert buffer.pop() == b"2"  # First item was overwritten
        assert buffer.pop() == b"3"
        assert buffer.pop() == b"4"

    def test_buffer_empty_msg(self):
        """Test buffer splits on marker."""
        buffer = CircularBuffer(10, element_size=5)
        data = b"\xff\xff\xff\x55"
        buffer.append(data)

        assert buffer.size == 0


class TestSerialInputAdapter:
    """Test SerialInputAdapter."""

    @patch("multilat_solver.input_adapters.serial.Serial")
    def test_serial_adapter_start_stop(self, mock_serial):
        """Test serial adapter start and stop."""
        mock_port = MagicMock()
        mock_port.in_waiting = 0
        mock_serial.return_value = mock_port

        adapter = SerialInputAdapter("/dev/ttyUSB0", 460800)
        callback = Mock()

        adapter.start(callback)
        assert adapter.is_running() is True

        adapter.stop()
        assert adapter.is_running() is False
        mock_port.close.assert_called_once()


class TestUDPInputAdapter:
    """Test UDPInputAdapter."""

    @patch("multilat_solver.input_adapters.socket.socket")
    def test_udp_adapter_json_dict(self, mock_socket_class):
        """Test UDP adapter with JSON dict format."""
        payload = json.dumps({"1": 2.5, "2": 3.1}).encode()
        call_count = [0]

        def recv_into(buffer):
            call_count[0] += 1
            if call_count[0] == 1:
                n = min(len(payload), len(buffer))
                buffer[:n] = payload[:n]
                return n
            raise socket.timeout  # subsequent calls: timeout so loop can check self.running

        mock_socket = MagicMock()
        mock_socket.recv_into.side_effect = recv_into

        mock_socket_class.return_value = mock_socket

        adapter = UDPInputAdapter("127.0.0.1", 5005)
        callback = Mock()

        adapter.start(callback)
        time.sleep(0.3)  # Let the read loop process one message
        adapter.stop()

        callback.assert_called_once()
        args = callback.call_args[0]
        assert args[0] == {1: 2.5, 2: 3.1}


class TestFileInputAdapter:
    """Test FileInputAdapter."""

    @patch("builtins.open", create=True)
    def test_file_adapter_json_dict(self, mock_open):
        """Test file adapter with JSON dict format."""
        mock_file = MagicMock()
        mock_file.__enter__.return_value = mock_file
        mock_file.__exit__.return_value = None
        mock_file.read.return_value = json.dumps({"1": 2.5, "2": 3.1})
        mock_open.return_value = mock_file

        adapter = FileInputAdapter("test.json", poll_interval=0.1)
        callback = Mock()

        adapter.start(callback)
        time.sleep(0.2)  # Wait for poll
        adapter.stop()

        # File should have been read
        mock_open.assert_called()


class TestRegisterInputAdapter:
    """Test register_input_adapter."""

    def test_register_input_adapter(self):
        """Test register_input_adapter."""

        class CustomInputAdapter(InputAdapter):
            """Custom input adapter."""

            def __init__(self, config: dict):
                self.config = config
                super().__init__(config)

            def start(self, callback: Callable[[Dict[int, Dict[str, float]]], None]):
                pass

            def stop(self):
                pass

            def is_running(self) -> bool:
                return False

        register_input_adapter("custom", CustomInputAdapter)
        assert "custom" in INPUT_ADAPTERS



"""Input adapters for receiving distance measurements from various sources."""

import json
import logging
import socket
import struct
import threading
import time
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, Optional

import serial

logger = logging.getLogger(__name__)


class InputAdapter(ABC):
    """Base class for input adapters."""

    @abstractmethod
    def start(self, callback: Callable[[Dict[int, Dict[str, float]]], None]):
        """Start receiving data and call callback with distance measurements.

        callback(distances: Dict[int, Dict[str, float]])
            Contains distance measurements for each anchor.
            distances[anchor_id] = {"distance": distance, "timestamp": timestamp}
        """
        pass

    @abstractmethod
    def stop(self):
        """Stop receiving data."""
        pass

    @abstractmethod
    def is_running(self) -> bool:
        """Check if adapter is running."""
        pass


class SerialInputAdapter(InputAdapter):
    """Serial port input adapter (from RTK-DW1000 implementation)."""

    def __init__(self, port: str, baud: int = 460800, timeout: float = 1.0):
        """Initialize serial input adapter."""
        self.port = port
        self.baud = baud
        self.timeout = timeout
        self.ser: serial.Serial | None = None
        self.running: bool = False
        self.callback: Callable[[Dict[int, Dict[str, float]]], None] | None = None
        self.thread: threading.Thread | None = None
        self.buffer: CircularBuffer = CircularBuffer(100, element_size=7)

    def start(self, callback: Callable[[Dict[int, Dict[str, float]]], None] | None = None):
        """Start reading from serial port."""
        if callback is not None:
            self.callback = callback
        try:
            self.ser = serial.Serial(
                self.port,
                self.baud,
                timeout=self.timeout,
                stopbits=serial.STOPBITS_ONE,
                bytesize=serial.EIGHTBITS,
                parity="E",
            )
            self.running = True
            self.thread = threading.Thread(target=self._read_loop, daemon=True)
            self.thread.start()
            logger.info("Serial adapter started on %s at %s baud", self.port, self.baud)
        except Exception as e:
            logger.error("Failed to connect to serial port: %s", e)
            raise

    def stop(self):
        """Stop reading from serial port."""
        self.running = False
        if self.ser:
            self.ser.close()
        if self.thread:
            self.thread.join(timeout=2.0)

    def is_running(self) -> bool:
        """Check if adapter is running."""
        return self.running

    def _read_loop(self):
        """Read from serial adapter."""
        ranges = {}
        while self.running:
            try:
                if self.ser.in_waiting > 0:
                    response = self.ser.read_all()
                    self.buffer.append(response)

                    while self.buffer.size > 0:
                        msg = self.buffer.pop()
                        if msg is None:
                            continue

                        message = Message(msg)
                        anchor_id = message.id
                        raw_val = message.data / 1000.0  # Convert mm to meters

                        if anchor_id is not None:
                            ranges[anchor_id] = {"distance": raw_val, "timestamp": time.time_ns() // 1000}

                    if ranges and self.callback:
                        self.callback(ranges.copy())
                        ranges.clear()

                time.sleep(0.01)  # Small delay to prevent CPU spinning
            except Exception as e:
                logger.error("Error in serial read loop: %s", e)
                time.sleep(0.1)


class MQTTInputAdapter(InputAdapter):
    """MQTT input adapter."""

    def __init__(self, broker: str, port: int = 1883, topic: str = "uwb/distances"):
        """Initialize MQTT input adapter."""
        import paho.mqtt.client as mqtt

        self.mqtt = mqtt
        self.broker = broker
        self.port = port
        self.topic = topic
        self.client: mqtt.Client | None = None
        self.running = False
        self.callback: Callable[[Dict[int, Dict[str, float]]], None] | None = None

    def start(self, callback: Callable[[Dict[int, Dict[str, float]]], None]):
        """Start MQTT subscriber."""
        try:
            self.callback = callback
            self.client = self.mqtt.Client()
            self.client.on_connect = self._on_connect
            self.client.on_message = self._on_message
            self.client.connect(self.broker, self.port, 60)
            self.running = True
            self.client.loop_start()
            logger.info(
                "MQTT adapter started on %s:%s, topic: %s",
                self.broker,
                self.port,
                self.topic,
            )
        except ImportError:
            raise ImportError("paho-mqtt is required for MQTT adapter. Install with: pip install paho-mqtt")
        except Exception as e:
            logger.error("Failed to connect to MQTT broker: %s", e)
            raise

    def stop(self):
        """Stop MQTT subscriber."""
        self.running = False
        if self.client:
            self.client.loop_stop()
            self.client.disconnect()

    def is_running(self) -> bool:
        """Check if adapter is running."""
        return self.running

    def _on_connect(self, client, userdata, flags, rc):
        """Handle MQTT connection callback."""
        del userdata, flags, rc
        client.subscribe(self.topic)

    def _on_message(self, client, userdata, msg):
        """MQTT message callback."""
        del client, userdata
        try:
            data = json.loads(msg.payload.decode())
            # Expected format: {"anchor_id": distance, ...} or [{"id": anchor_id, "distance": dist}, ...]
            if isinstance(data, dict):
                distances = {int(k): float(v) for k, v in data.items()}
            elif isinstance(data, list):
                distances = {int(item["id"]): float(item["distance"]) for item in data}
            else:
                logger.warning("Unexpected MQTT message format: %s", data)
                return

            if self.callback:
                self.callback(distances)
        except self.mqtt.exceptions.MqttError as e:
            logger.error("Error processing MQTT message: %s", e)


class UDPInputAdapter(InputAdapter):
    """UDP input adapter."""

    def __init__(self, host: str = "0.0.0.0", port: int = 5005):
        """Initialize UDP input adapter."""
        self.host = host
        self.port = port
        self.socket: socket.socket | None = None
        self.running = False
        self.callback: Callable[[Dict[int, Dict[str, float]]], None] | None = None
        self.thread: threading.Thread | None = None
        self.buffer = bytearray(1024)

    def start(self, callback: Callable[[Dict[int, Dict[str, float]]], None]):
        """Start UDP listener."""
        if callback is not None:
            self.callback = callback
        try:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            self.socket.bind((self.host, self.port))
            self.socket.setblocking(False)
            # self.ready = select.select([self.socket], [], [], 0.1)  # timeout
            self.socket.settimeout(1.0)
            self.running = True
            self.thread = threading.Thread(target=self._read_loop, daemon=True)
            self.thread.start()
            logger.info("UDP adapter started on %s:%s", self.host, self.port)
        except Exception as e:
            logger.error("Failed to start UDP adapter: %s", e)
            raise

    def stop(self):
        """Stop UDP listener."""
        self.running = False
        if self.socket:
            self.socket.close()
        if self.thread:
            self.thread.join(timeout=1.0)

    def is_running(self) -> bool:
        """Check if adapter is running."""
        return self.running

    def _read_loop(self):
        """Read from UDP adapter."""
        while self.running:
            if self.socket is None:
                self.running = False
                raise RuntimeError("Socket not initialized")

            data = self.socket.recv_into(self.buffer)
            if not data:
                self.running = False
                raise RuntimeError("Socket connection closed")
            if data == 0:
                raise RuntimeError("No data recieved")
            try:
                message = json.loads(self.buffer[:data].decode())
                # Expected format: {"anchor_id": distance, ...} or [{"id": anchor_id, "distance": dist}, ...]
                if isinstance(message, dict):
                    distances = {int(k): float(v) for k, v in message.items()}
                elif isinstance(message, list):
                    distances = {int(item["id"]): float(item["distance"]) for item in message}
                else:
                    continue

                if self.callback:
                    self.callback(distances)
                    return
            except json.JSONDecodeError as e:
                logger.warning("Invalid JSON in UDP message: %s", e)


class FileInputAdapter(InputAdapter):
    """File input adapter for reading from JSON file."""

    def __init__(self, filepath: str, poll_interval: float = 1.0):
        """Initialize file input adapter."""
        self.filepath = filepath
        self.poll_interval = poll_interval
        self.running = False
        self.callback: Callable[[Dict[int, Dict[str, float]]], None] | None = None
        self.thread: threading.Thread | None = None

    def start(self, callback: Callable[[Dict[int, Dict[str, float]]], None] | None = None):
        """Start reading from file."""
        if callback is not None:
            self.callback = callback
        self.running = True
        self.thread = threading.Thread(target=self._read_loop, daemon=True)
        self.thread.start()
        logger.info("File adapter started reading from %s", self.filepath)

    def stop(self):
        """Stop reading from file."""
        self.running = False
        if self.thread:
            self.thread.join(timeout=2.0)

    def is_running(self) -> bool:
        """Check if adapter is running."""
        return self.running

    def _read_loop(self):
        """Read from file adapter."""
        while self.running:
            try:
                with open(self.filepath, "r") as f:
                    data = json.load(f)
                    if isinstance(data, dict):
                        distances = {int(k): float(v) for k, v in data.items()}
                    elif isinstance(data, list):
                        distances = {int(item["id"]): float(item["distance"]) for item in data}
                    else:
                        continue

                    if self.callback:
                        self.callback(distances)
            except FileNotFoundError:
                logger.warning("File not found: %s", self.filepath)
            except json.JSONDecodeError as e:
                logger.warning("Invalid JSON in file: %s", e)
            except Exception as e:
                logger.error("Error reading file: %s", e)

            time.sleep(self.poll_interval)


# Helper classes from RTK-DW1000
class Message:
    """Message parser for serial protocol."""

    def __init__(self, data: bytes):
        """Initialize message parser."""
        self.id, self.data = struct.unpack("<BI", data)

    def __str__(self):
        """Return string representation of message."""
        return f"id: {self.id}, data: {self.data}"


class CircularBuffer:
    """Circular buffer for serial message parsing."""

    def __init__(self, size: int, element_size: int = 5):
        """Initialize circular buffer."""
        self.buffer = [b"\x00" for i in range(size)]
        self.head = 0
        self.size = 0
        self.tail = 0
        self.element_size = element_size

    def append(self, item: Optional[bytes]):
        """Append item to buffer."""
        if item is None:
            return
        items = item.split(b"\xff\xff\xff\x55")
        for item in items:
            if len(item) == 0:
                continue
            if len(item) > self.element_size:
                continue
            self.buffer[self.head] = item
            self.head = (self.head + 1) % len(self.buffer)
            self.size += 1
        if self.size > len(self.buffer):
            self.size = len(self.buffer)
            self.tail += 1

    def pop(self) -> Optional[bytes]:
        """Pop item from buffer."""
        if self.size == 0:
            return None
        item = self.buffer[self.tail]
        self.tail = (self.tail + 1) % len(self.buffer)
        self.size -= 1
        return item


INPUT_ADAPTERS = {
    "serial": SerialInputAdapter,
    "mqtt": MQTTInputAdapter,
    "udp": UDPInputAdapter,
    "file": FileInputAdapter,
}


def register_input_adapter(adapter_type: str, adapter: Any) -> None:
    """Register input adapter."""
    if adapter is None:
        raise ValueError("Adapter cannot be None")
    if not issubclass(adapter, InputAdapter):
        raise ValueError("Adapter must be a subclass of InputAdapter")
    if adapter_type in INPUT_ADAPTERS:
        raise ValueError(f"Adapter type {adapter_type} already registered")
    INPUT_ADAPTERS[adapter_type] = adapter



"""Common types for the MultiLat Localization System."""

from enum import Enum


class CalibrationType(str, Enum):
    """Calibration type enumeration."""

    NONE = "none"
    LINEAR = "linear"
    QUADRATIC = "quadratic"
    CUBIC = "cubic"


class MessageType(str, Enum):
    """Output message type enumeration."""

    GPS = "gps"
    POSITION = "position"
    BOTH = "both"



"""
Output adapters for sending position data to various destinations.
"""

import csv
import json
import logging
import socket
import time
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple

from multilat_solver.common_types import MessageType

logger = logging.getLogger(__name__)


class OutputAdapter(ABC):
    """Base class for output adapters."""

    def __init__(self, message_type: MessageType, frequency: int):
        """Initialize output adapter."""
        self._message_type = message_type
        self._frequency = frequency
        self._sleep_time = 1.0 / frequency if frequency > 0 else 0.1
        self._last_send_position_time = 0.0
        self._last_send_gps_time = 0.0

    def send_position(self, position: Tuple[float, float, float], timestamp: Optional[float] = None):
        """Check if position should be sent based on message type and frequency throttling."""
        if self._message_type not in [MessageType.POSITION, MessageType.BOTH]:
            return False
        if time.time() - self._last_send_position_time < self._sleep_time:
            return False
        self._last_send_position_time = time.time()
        return True

    def send_gps(
        self,
        coordinates: Tuple[float, float, float],
        timestamp: Optional[float] = None,
        n_satellites: int = 10,
    ):
        """Check if GPS should be sent based on message type and frequency throttling.

        :param coordinates: Tuple[float, float, float] - latitude, longitude, altitude in degrees and meters
        :param timestamp: Optional[float] - timestamp in seconds
        :param n_satellites: Number of satellites
        """
        if self._message_type not in [MessageType.GPS, MessageType.BOTH]:
            return False
        if time.time() - self._last_send_gps_time < self._sleep_time:
            return False
        self._last_send_gps_time = time.time()
        return True

    @abstractmethod
    def close(self):
        """Close the adapter."""
        pass


class MavlinkOutputAdapter(OutputAdapter):
    """MAVLink output adapter using pymavlink."""

    def __init__(
        self,
        connection_string: str = "udpout:127.0.0.1:14550",
        system_id: int = 1,
        component_id: int = 1,
        message_type: MessageType = MessageType.BOTH,
        frequency: int = 10,
        reconnect_attempts: int = 3,
        heartbeat_wait_s: float = 10.0,
    ):
        """
        Initialize MAVLink adapter.

        :param connection_string: MAVLink connection string
                (e.g., "udpout:127.0.0.1:14550", "tcp:127.0.0.1:5760")
        :param system_id: MAVLink system ID
        :param component_id: MAVLink component ID
        :param message_type: Message type
        :param frequency: Frequency in Hz
        """
        logger.info(
            "MavlinkOutputAdapter initialized: %s, %s, %s, %s",
            connection_string,
            system_id,
            component_id,
            frequency,
        )
        self.connection = None
        self.reconnect_attempts = reconnect_attempts
        self.heartbeat_wait_s = heartbeat_wait_s
        self.last_heartbeat_time = time.monotonic()
        self.component_id = component_id
        self.system_id = system_id
        self.connection_string = connection_string
        self.connected = False

        self._try_reconnect()
        super().__init__(message_type, frequency)

    def _connect(self):
        from pymavlink import mavutil

        return mavutil.mavlink_connection(
            self.connection_string, source_system=self.system_id, source_component=self.component_id
        )

    def _try_reconnect(self):
        attempts = self.reconnect_attempts
        while attempts > 0:
            if self.connection is not None:
                self.connection.close()
            self.connection = self._connect()
            res = self.connection.wait_heartbeat(timeout=self.heartbeat_wait_s)
            if res is not None:
                self.last_heartbeat_time = time.monotonic()
                self.connected = True
                return
            attempts -= 1
            print(f"MavlinkOutputAdapter heartbeat timeout. Attempts left: {attempts}")
        self.connected = False

    def wait_heartbeat(self):
        """Wait for heartbeat from PX4."""
        while not self.connected:
            time.sleep(0.2)
            self._try_reconnect()

    def send_position(self, position: Tuple[float, float, float], timestamp: Optional[float] = None):
        """Send position via MAVLink LOCAL_POSITION_NED message."""
        if not self.connected:
            time.sleep(0.2)
            self._try_reconnect()
            return
        if time.monotonic() - self.last_heartbeat_time > self.heartbeat_wait_s:
            self.connected = False
            self._try_reconnect()
            return

        if not super().send_position(position, timestamp):
            return

        x, y, z = position
        timestamp_ms = int(timestamp) if timestamp else int(time.time_ns() / 1000)
        # Send LOCAL_POSITION_NED message
        self.connection.mav.local_position_ned_send(
            timestamp_ms % 4294967295,
            x,  # x position (m)
            y,  # y position (m)
            z,  # z position (m)
            0,  # vx velocity (m/s)
            0,  # vy velocity (m/s)
            0,  # vz velocity (m/s),
        )
        logger.info(f"Mavlink Local position send: {(x, y, z)}")

    def send_gps(
        self,
        coordinates: Tuple[float, float, float],
        timestamp: Optional[float] = None,
        n_satellites: int = 10,
    ):
        """Send GPS coordinates.

        :param coordinates: Tuple[float, float, float] - latitude, longitude, altitude in degrees and meters
        :param timestamp: Optional[float] - timestamp in seconds
        :param n_satellites: Number of satellites
        """
        if not self.connected:
            time.sleep(0.2)
            self._try_reconnect()
            return
        if time.monotonic() - self.last_heartbeat_time > self.heartbeat_wait_s:
            self.connected = False
            self._try_reconnect()
            return
        if not super().send_gps(coordinates, timestamp):
            return

        lat, lon, alt = coordinates
        timestamp_ms = int(timestamp) if timestamp else int(time.time_ns() // 1000)

        vn = ve = vd = 0  # cm/s (NED)
        vel = 0  # ground speed cm/s
        cog = 0  # course over ground cdeg (0..35999)

        self.connection.mav.hil_gps_send(
            timestamp_ms,  # time_usec (microseconds)
            3,  # fix_type (3 = GPS_FIX_TYPE_3D_FIX)
            int(lat * 1e7),  # lat (degrees * 1e7)
            int(lon * 1e7),  # lon (degrees * 1e7)
            int(alt * 1000),  # alt (mm above sea level)
            80,  # eph (cm) - horizontal position uncertainty
            120,  # epv (cm) - vertical position uncertainty
            vel,  # vel (cm/s) - ground speed
            vn,  # vn (cm/s) - velocity north
            ve,  # ve (cm/s) - velocity east
            vd,  # vd (cm/s) - velocity down
            cog,  # cog (cm/s) - Course over ground (NOT heading, but direction of movement),
            # 0.0..359.99 degrees. If unknown, set to: UINT16_MAX
            n_satellites,  # satellites_visible
        )
        logger.info(f"Mavlink GPS send: {lat, lon, alt}")

    def close(self):
        """Close MAVLink connection."""
        if self.connection:
            self.connection.close()


class MavrosOutputAdapter(OutputAdapter):
    """MAVROS output adapter (ROS2)."""

    def __init__(
        self,
        topic_prefix: str = "mavros",
        frame_id: str = "map",
        frequency: int = 10,
        message_type: MessageType = MessageType.BOTH,
    ):
        """
        Initialize MAVROS adapter.

        :param topic_prefix: ROS2 topic prefix (default: "mavros")
        :param frame_id: Frame ID for messages
        """
        self.topic_prefix = topic_prefix
        self.frame_id = frame_id
        self.node = None
        self.pose_pub = None
        self.fake_gps_pub = None
        self._initialize()
        super().__init__(message_type, frequency)

    def _initialize(self):
        """Initialize ROS2 node and publisher."""
        try:
            # ignore import error
            import rclpy  # noqa: F401
            from geometry_msgs.msg import PoseStamped
            from sensor_msgs.msg import NavSatFix

            if not rclpy.ok():
                rclpy.init()

            from rclpy.node import Node

            self.node = Node("multilat_solver_output")
            self.PoseStamped = PoseStamped
            self.NavSatFix = NavSatFix

            topic = f"{self.topic_prefix}/fake_gps/mocap/pose"
            self.pose_pub = self.node.create_publisher(PoseStamped, topic, self.frequency)
            self.fake_gps_pub = self.node.create_publisher(NavSatFix, "uwb/gps", self.frequency)

            logger.info("MAVROS adapter initialized: %s", topic)
        except ImportError:
            raise ImportError("rclpy is required for MAVROS adapter. Install ROS2 dependencies.")

    def _spin_once(self):
        """Spin once to process callbacks."""
        if self.node:
            import rclpy

            rclpy.spin_once(self.node, timeout_sec=0.01)

    def send_position(self, position: Tuple[float, float, float], timestamp: Optional[float] = None):
        """Send position via ROS2 PoseStamped message."""
        if not super().send_position(position, timestamp):
            return
        if self.pose_pub is None:
            return

        x, y, z = position

        pose = self.PoseStamped()
        if self.node:
            pose.header.stamp = self.node.get_clock().now().to_msg()
        pose.header.frame_id = self.frame_id
        pose.pose.position.x = x
        pose.pose.position.y = y
        pose.pose.position.z = z
        pose.pose.orientation.w = 1.0

        self.pose_pub.publish(pose)

        # Spin once to process callbacks
        if self.node:
            self._spin_once()

    def send_gps(
        self,
        coordinates: Tuple[float, float, float],
        timestamp: Optional[float] = None,
        n_satellites: int = 10,
    ):
        """Send GPS coordinates.

        :param coordinates: Tuple[float, float, float] - latitude, longitude, altitude in degrees and meters
        :param timestamp: Optional[float] - timestamp in seconds
        :param n_satellites: Number of satellites
        """
        if not super().send_gps(coordinates, timestamp, n_satellites):
            return
        if self.fake_gps_pub is None:
            return
        nav = self.NavSatFix()
        if self.node:
            nav.header.stamp = self.node.get_clock().now().to_msg()
        nav.header.frame_id = self.frame_id
        nav.latitude = coordinates[0]
        nav.longitude = coordinates[1]
        nav.altitude = coordinates[2]
        self.fake_gps_pub.publish(nav)

        # Spin once to process callbacks
        if self.node:
            import rclpy

            rclpy.spin_once(self.node, timeout_sec=0.01)

    def close(self):
        """Close ROS2 node."""
        if self.node:
            try:
                import rclpy

                rclpy.shutdown()
            except ImportError as exc:
                raise ImportError(
                    "rclpy is required for MAVROS adapter. Install ROS2 dependencies."
                ) from exc  # noqa: B904


class UDPOutputAdapter(OutputAdapter):
    """UDP output adapter."""

    def __init__(
        self,
        host: str = "127.0.0.1",
        port: int = 5006,
        message_type: MessageType = MessageType.BOTH,
        frequency: int = 10,
    ):
        """
        Initialize UDP adapter.

        :param host: Target host
        :param port: Target port
        :param message_type: Message type
        :param frequency: Frequency in Hz
        """
        self.host = host
        self.port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        logger.info("UDP output adapter initialized: %s:%s", host, port)
        super().__init__(message_type, frequency)

    def send_position(self, position: Tuple[float, float, float], timestamp: Optional[float] = None):
        """Send position via UDP as JSON.

        :param position: Tuple[float, float, float] - x, y, z in meters
        :param timestamp: Optional[float] - timestamp in seconds
        """
        if not super().send_position(position, timestamp):
            return
        x, y, z = position
        message = {
            "position": {"x": x, "y": y, "z": z},
            "timestamp": timestamp or time.time(),
        }

        data = json.dumps(message).encode()
        self.socket.sendto(data, (self.host, self.port))

    def send_gps(
        self,
        coordinates: Tuple[float, float, float],
        timestamp: Optional[float] = None,
        n_satellites: int = 10,
    ):
        """Send GPS coordinates.

        :param coordinates: Tuple[float, float, float] - latitude, longitude, altitude in degrees and meters
        :param timestamp: Optional[float] - timestamp in seconds
        :param n_satellites: Number of satellites
        """
        if not super().send_gps(coordinates, timestamp, n_satellites):
            return
        if self.socket is None:
            return
        lat, lon, alt = coordinates
        message = {
            "latitude": lat,
            "longitude": lon,
            "altitude": alt,
            "timestamp": timestamp or time.time(),
            "satellites": n_satellites,
        }
        data = json.dumps(message).encode()
        self.socket.sendto(data, (self.host, self.port))

    def close(self):
        """Close UDP socket."""
        if self.socket:
            self.socket.close()


class FileOutputAdapter(OutputAdapter):
    """File output adapter for logging positions."""

    def __init__(
        self,
        filepath: str | None = None,
        mode: str | None = None,
        message_type: MessageType = MessageType.BOTH,
        frequency: int = 10,
        **kwargs,
    ):
        """
        Initialize file adapter.

        :param filepath: Path to output file
        :param mode: File mode ('a' for append, 'w' for write)
        :param message_type: Message type
        :param frequency: Frequency in Hz
        :param kwargs: Additional keyword arguments
        """
        if filepath is None:
            filepath = kwargs.get("filepath", "positions.jsonl")
        extention = filepath.split(".")[-1]

        if extention not in ["jsonl", "csv"]:
            raise ValueError(f"Invalid file extension: {extention}")

        filename = filepath.replace(extention, "")[:-1]

        if mode is None:
            mode = kwargs.get("mode", "a")
        self.files = {}
        self.writers = {}
        self.is_csv = extention == "csv"

        if message_type in [MessageType.POSITION, MessageType.BOTH]:
            self.files[MessageType.POSITION] = open(f"{filename}_positions.{extention}", mode)
            if self.is_csv:
                self.writers[MessageType.POSITION] = csv.writer(self.files[MessageType.POSITION])
                self.writers[MessageType.POSITION].writerow(["timestamp", "x", "y", "z"])

        if message_type in [MessageType.GPS, MessageType.BOTH]:
            self.files[MessageType.GPS] = open(f"{filename}_gps.{extention}", mode)
            if self.is_csv:
                self.writers[MessageType.GPS] = csv.writer(self.files[MessageType.GPS])
                self.writers[MessageType.GPS].writerow(
                    ["timestamp", "latitude", "longitude", "altitude", "satellites"]
                )

        logger.info("File output adapter initialized: %s", filepath)
        super().__init__(message_type, frequency)

    def send_position(self, position: Tuple[float, float, float], timestamp: Optional[float] = None):
        """Write position to file as JSON line.

        :param position: Tuple[float, float, float] - x, y, z in meters
        :param timestamp: Optional[float] - timestamp in seconds
        """
        if not super().send_position(position, timestamp):
            return
        if self.files[MessageType.POSITION] is None:
            return
        x, y, z = position
        message = {
            "position": {"x": x, "y": y, "z": z},
            "timestamp": timestamp or time.time(),
        }
        if self.is_csv:
            self.writers[MessageType.POSITION].writerow([timestamp or time.time(), x, y, z])
        else:
            self.files[MessageType.POSITION].write(json.dumps(message) + "\n")
        self.files[MessageType.POSITION].flush()

    def send_gps(
        self,
        coordinates: Tuple[float, float, float],
        timestamp: Optional[float] = None,
        n_satellites: int = 10,
    ):
        """Send GPS coordinates.

        :param coordinates: Tuple[float, float, float] - latitude, longitude, altitude in degrees and meters
        :param timestamp: Optional[float] - timestamp in seconds
        :param n_satellites: Number of satellites
        """
        if not super().send_gps(coordinates, timestamp):
            return

        if self.files[MessageType.GPS] is None:
            return
        lat, lon, alt = coordinates
        message = {
            "latitude": lat,
            "longitude": lon,
            "altitude": alt,
            "timestamp": timestamp or time.time(),
            "satellites": n_satellites,
        }
        if self.is_csv:
            self.writers[MessageType.GPS].writerow([timestamp or time.time(), lat, lon, alt, n_satellites])
        else:
            self.files[MessageType.GPS].write(json.dumps(message) + "\n")
        self.files[MessageType.GPS].flush()

    def close(self):
        """Close file."""
        for file in self.files.values():
            file.close()


class ConsoleOutputAdapter(OutputAdapter):
    """Console output adapter for debugging."""

    def __init__(
        self,
        format: str | None = None,
        message_type: MessageType = MessageType.BOTH,
        frequency: int = 10,
        **kwargs,
    ):
        """
        Initialize console adapter.

        :param format: Output format ('json' or 'human')
        :param message_type: Message type
        :param frequency: Frequency in Hz
        :param kwargs: Additional keyword arguments
        """
        if format is None:
            format = kwargs.get("format", "json")
        self.format = format
        logger.info("Console output adapter initialized")
        super().__init__(message_type, frequency)

    def send_position(self, position: Tuple[float, float, float], timestamp: Optional[float] = None):
        """Print position to console (stdout) for debugging.

        :param position: Tuple[float, float, float] - x, y, z in meters
        :param timestamp: Optional[float] - timestamp in seconds
        """
        if not super().send_position(position, timestamp):
            return
        x, y, z = position
        if self.format == "json":
            message = {
                "position": {"x": x, "y": y, "z": z},
                "timestamp": timestamp or time.time(),
            }
            print(json.dumps(message))
        else:
            print(f"Position: x={x:.3f}, y={y:.3f}, z={z:.3f}")

    def send_gps(
        self,
        coordinates: Tuple[float, float, float],
        timestamp: Optional[float] = None,
        n_satellites: int = 10,
    ):
        """Send GPS coordinates to console (stdout) for debugging.

        :param coordinates: Tuple[float, float, float] - latitude, longitude, altitude in degrees and meters
        :param timestamp: Optional[float] - timestamp in seconds
        :param n_satellites: Number of satellites
        """
        if not super().send_gps(coordinates, timestamp):
            return
        if self.format == "json":
            message = {
                "latitude": coordinates[0],
                "longitude": coordinates[1],
                "altitude": coordinates[2],
                "timestamp": timestamp or time.time(),
                "satellites": n_satellites,
            }
            print(json.dumps(message))
        else:
            print(
                f"GPS: latitude={coordinates[0]:.6f},\
                longitude={coordinates[1]:.6f},\
                altitude={coordinates[2]:.3f},\
                satellites={n_satellites}"
            )

    def close(self):
        """Nothing to close for console output."""
        pass


class MultiOutputAdapter:
    """Adapter that sends to multiple output adapters."""

    def __init__(self, adapters: list[OutputAdapter]):
        """
        Initialize multi-output adapter.

        :param adapters: List of output adapters
        """
        self.adapters: list[OutputAdapter] = adapters
        logger.info("Multi-output adapter initialized with %s adapters", len(adapters))

    def send_position(self, position: Tuple[float, float, float], timestamp: Optional[float] = None):
        """Send position to all adapters.

        :param position: Tuple[float, float, float] - x, y, z in meters
        :param timestamp: Optional[float] - timestamp in seconds
        """
        for adapter in self.adapters:
            adapter.send_position(position, timestamp)

    def send_gps(
        self,
        coordinates: Tuple[float, float, float],
        timestamp: Optional[float] = None,
        n_satellites: int = 10,
    ):
        """Send GPS coordinates to all adapters.

        :param coordinates: Tuple[float, float, float] - latitude, longitude, altitude in degrees and meters
        :param timestamp: Optional[float] - timestamp in seconds
        :param n_satellites: Number of satellites
        """
        for adapter in self.adapters:
            adapter.send_gps(coordinates, timestamp, n_satellites)

    def close(self):
        """Close all adapters."""
        for adapter in self.adapters:
            adapter.close()


OUTPUT_ADAPTERS = {
    "mavlink": MavlinkOutputAdapter,
    "mavros": MavrosOutputAdapter,
    "udp": UDPOutputAdapter,
    "file": FileOutputAdapter,
    "console": ConsoleOutputAdapter,
    "multi": MultiOutputAdapter,
}


def register_output_adapter(adapter_type: str, adapter: Any) -> None:
    """Register output adapter."""
    if adapter_type not in OUTPUT_ADAPTERS:
        raise ValueError(f"Invalid adapter type: {adapter_type}")
    if adapter is None:
        raise ValueError("Adapter cannot be None")
    if not issubclass(type(adapter), OutputAdapter):
        raise ValueError("Adapter must be an instance of OutputAdapter")
    if adapter_type in OUTPUT_ADAPTERS:
        raise ValueError(f"Adapter type {adapter_type} already registered")
    OUTPUT_ADAPTERS[adapter_type] = adapter



"""
Unit tests for output adapters.
"""

import json
import time
from unittest.mock import MagicMock, Mock, patch

from multilat_solver.output_adapters import (
    ConsoleOutputAdapter,
    FileOutputAdapter,
    MavlinkOutputAdapter,
    MavrosOutputAdapter,
    MessageType,
    MultiOutputAdapter,
    UDPOutputAdapter,
)


class TestConsoleOutputAdapter:
    """Test ConsoleOutputAdapter."""

    def test_console_output_human(self, capsys):
        """Test console output in human format."""
        adapter = ConsoleOutputAdapter(format="human")
        adapter.send_position((1.0, 2.0, 3.0))

        captured = capsys.readouterr()
        assert "x=1.000" in captured.out
        assert "y=2.000" in captured.out
        assert "z=3.000" in captured.out

    def test_console_output_json(self, capsys):
        """Test console output in JSON format."""
        adapter = ConsoleOutputAdapter(format="json")
        adapter.send_position((1.0, 2.0, 3.0), timestamp=1234567890.0)

        captured = capsys.readouterr()
        data = json.loads(captured.out.strip())
        assert data["position"]["x"] == 1.0
        assert data["position"]["y"] == 2.0
        assert data["position"]["z"] == 3.0
        assert data["timestamp"] == 1234567890.0


class TestUDPOutputAdapter:
    """Test UDPOutputAdapter."""

    @patch("multilat_solver.output_adapters.socket.socket")
    def test_udp_output(self, mock_socket_class):
        """Test UDP output sends JSON."""
        mock_socket = MagicMock()
        mock_socket_class.return_value = mock_socket

        adapter = UDPOutputAdapter("127.0.0.1", 5006)
        adapter.send_position((1.0, 2.0, 3.0), timestamp=1234567890.0)

        # Verify socket was created
        mock_socket_class.assert_called_once()

        # Verify sendto was called
        assert mock_socket.sendto.called

        # Check the sent data
        call_args = mock_socket.sendto.call_args
        sent_data = json.loads(call_args[0][0].decode())
        assert sent_data["position"]["x"] == 1.0
        assert sent_data["position"]["y"] == 2.0
        assert sent_data["position"]["z"] == 3.0


class TestFileOutputAdapter:
    """Test FileOutputAdapter."""

    @patch("builtins.open", create=True)
    def test_file_output_append(self, mock_open):
        """Test file output appends JSON lines."""
        mock_file = MagicMock()
        mock_open.return_value = mock_file

        adapter = FileOutputAdapter("test.jsonl", mode="a", message_type=MessageType.POSITION)
        adapter.send_position((1.0, 2.0, 3.0), timestamp=time.time() - 10)

        # Verify file was opened in append mode
        mock_open.assert_called_with("test_positions.jsonl", "a")

        # Verify write was called
        assert mock_file.write.called
        assert mock_file.flush.called

        # Check written data
        written_data = mock_file.write.call_args[0][0]
        data = json.loads(written_data.strip())
        assert data["position"]["x"] == 1.0

    def test_file_output_close(self):
        """Test file output closes properly."""
        adapter = FileOutputAdapter("test.jsonl", mode="a")
        adapter.close()

        # File should be closed (would need proper mock to verify)


class TestMavlinkOutputAdapter:
    """Test MavlinkOutputAdapter."""

    def test_mavlink_output(self):
        """Test MAVLink output sends LOCAL_POSITION_NED."""
        # Mock the _initialize method to prevent actual socket creation
        with patch.object(MavlinkOutputAdapter, "_try_reconnect", return_value=None):
            adapter = MavlinkOutputAdapter("udpout:127.0.0.1:14550")
            # Mock the mav object and its methods
            adapter.connection = MagicMock()
            adapter.connection.mav = MagicMock()
            adapter.last_heartbeat_time = time.time()
            adapter.connected = True
            adapter.send_position((1.0, 2.0, 3.0))
            assert adapter.connection.mav.local_position_ned_send.called


class TestMavrosOutputAdapter:
    """Test MavrosOutputAdapter."""

    def test_mavros_output(self):
        """Test MAVROS output publishes PoseStamped."""
        # Mock ROS2 components
        mock_node = MagicMock()
        mock_clock = MagicMock()
        mock_time_msg = MagicMock()
        mock_clock.now.return_value.to_msg.return_value = mock_time_msg
        mock_node.get_clock.return_value = mock_clock

        mock_publisher = MagicMock()
        mock_node.create_publisher.return_value = mock_publisher
        mock_pose = MagicMock()

        # Patch _initialize to prevent actual ROS2 initialization
        with patch.object(MavrosOutputAdapter, "_initialize", return_value=None):
            adapter = MavrosOutputAdapter()
            adapter.node = mock_node
            adapter.pose_pub = mock_publisher
            adapter.fake_gps_pub = mock_publisher
            adapter._spin_once = MagicMock()
            adapter.PoseStamped = MagicMock(return_value=mock_pose)

            adapter.send_position((1.0, 2.0, 3.0))
            # Check spin_once was called
            assert adapter._spin_once.called
            # Verify publisher was created
            assert adapter.pose_pub.publish.called


class TestMultiOutputAdapter:
    """Test MultiOutputAdapter."""

    def test_multi_output_sends_to_all(self):
        """Test multi-output sends to all adapters."""
        adapter1 = Mock()
        adapter2 = Mock()
        adapter3 = Mock()

        multi = MultiOutputAdapter([adapter1, adapter2, adapter3])
        multi.send_position((1.0, 2.0, 3.0))

        # Verify all adapters received the position
        adapter1.send_position.assert_called_once_with((1.0, 2.0, 3.0), None)
        adapter2.send_position.assert_called_once_with((1.0, 2.0, 3.0), None)
        adapter3.send_position.assert_called_once_with((1.0, 2.0, 3.0), None)

    def test_multi_output_handles_errors(self):
        """Test multi-output handles adapter errors gracefully."""
        adapter1 = Mock()
        adapter2 = Mock(side_effect=Exception("Test error"))
        adapter3 = Mock()

        multi = MultiOutputAdapter([adapter1, adapter2, adapter3])

        # Should not raise, but continue to other adapters
        multi.send_position((1.0, 2.0, 3.0))

        adapter1.send_position.assert_called_once()
        adapter3.send_position.assert_called_once()

    def test_multi_output_close(self):
        """Test multi-output closes all adapters."""
        adapter1 = Mock()
        adapter2 = Mock()

        multi = MultiOutputAdapter([adapter1, adapter2])
        multi.close()

        adapter1.close.assert_called_once()
        adapter2.close.assert_called_once()



"""Unit tests for config module."""

from typing import Callable, Dict, Optional, Tuple

import pytest

from multilat_solver.config import ConfigBuilder, LocalizationConfig, MessageType
from multilat_solver.input_adapters import InputAdapter
from multilat_solver.output_adapters import ConsoleOutputAdapter, OutputAdapter


class TestConfigBuilder:
    """Test ConfigBuilder."""

    @pytest.fixture
    def config_builder(self):
        """Return a ConfigBuilder instance."""
        return ConfigBuilder()

    @pytest.fixture
    def localization_config(self):
        """Return a LocalizationConfig instance."""
        return LocalizationConfig(
            anchor_positions=[[0, 0, 0], [1, 0, 0], [0, 1, 0]],
            origin_coordinates=[0, 0, 0],
        )

    def test_without_settings(self, config_builder):
        """Test ConfigBuilder without settings."""
        with pytest.raises(ValueError):
            config_builder.build()

    def test_without_input(self, config_builder, localization_config):
        """Test ConfigBuilder without input."""
        config_builder.with_output(
            output_str="console",
            adapter=ConsoleOutputAdapter(format="human"),
        )
        config_builder.with_localization(localization_config)
        with pytest.raises(ValueError, match="Input configuration is required"):
            config_builder.build()

    def test_without_output(self, config_builder, localization_config):
        """Test ConfigBuilder without output."""
        config_builder.with_file_input(filepath="test.txt", poll_interval=1.0)
        config_builder.with_localization(localization_config)
        with pytest.raises(ValueError, match="Output configuration is required"):
            config_builder.build()

    def test_without_localization(self, config_builder):
        """Test ConfigBuilder without localization type."""
        config_builder.with_file_input(filepath="test.txt", poll_interval=1.0)
        config_builder.with_output(
            output_str="mavlink",
            adapter=ConsoleOutputAdapter(message_type=MessageType.POSITION, frequency=10),
        )
        with pytest.raises(ValueError, match="Localization configuration is required"):
            config_builder.build()

    def test_custom_output_adapter(self, config_builder, localization_config):
        """Test custom output adapter."""

        class CustomOutputAdapter(OutputAdapter):
            """Custom output adapter."""

            def __init__(
                self,
                custom_config: str,
                message_type: MessageType = MessageType.POSITION,
                frequency: int = 10,
            ):
                self.custom_config = custom_config
                super().__init__(message_type, frequency)

            def send_position(
                self,
                position: Tuple[float, float, float],
                timestamp: Optional[float] = None,
            ):
                """Send position."""
                print(f"Position: {position}")

            def send_gps(
                self,
                coordinates: Tuple[float, float, float],
                timestamp: Optional[float] = None,
            ):
                """Send GPS coordinates."""
                print(f"GPS coordinates: {coordinates}")

            def close(self):
                """Close adapter."""
                print("Closing custom output adapter.")

        config_builder.with_localization(localization_config)
        config_builder.with_file_input(filepath="test.txt", poll_interval=1.0)
        config_builder.with_output(
            output_str="custom",
            adapter=CustomOutputAdapter(custom_config="custom_config"),
        )
        config = config_builder.build()
        assert config.output.types[0] == "custom"
        assert config.output.adapters["custom"] is not None

    def test_register_input_adapter_with_config(self, config_builder, localization_config):
        """Test register_input_adapter with config."""

        class CustomInputAdapter(InputAdapter):
            """Custom input adapter."""

            def __init__(self, custom_param: int):
                self.custom_param = custom_param

            def start(self, callback: Callable[[Dict[int, Dict[str, float]]], None]):
                pass

            def stop(self):
                pass

            def is_running(self) -> bool:
                return False

            def _read_loop(self):
                pass

        config_builder.with_output(
            output_str="console",
            adapter=ConsoleOutputAdapter(format="human"),
        )
        config_builder.with_localization(localization_config)
        config_builder.with_input(input_str="custom", adapter=CustomInputAdapter(custom_param=1))
        config = config_builder.build()
        assert config.input.type_name == "custom"
        assert config.input.adapter is not None
        assert config.input.adapter.custom_param == 1



"""Core localization module using multilateration algorithm."""

import logging
from typing import Dict, List, Optional, Tuple

import localization as lx
import pymap3d as pm

from multilat_solver.config import LocalizationConfig

logger = logging.getLogger(__name__)


def calibrated_linear(k: float, b: float, x: float) -> float:
    """Linear calibration: k * x + b."""
    return k * x + b


def calibrated_quadratic(a: float, b: float, c: float, x: float) -> float:
    """Quadratic calibration: a * x^2 + b * x + c."""
    return a * x**2 + b * x + c


def calibrated_cubic(a: float, b: float, c: float, d: float, x: float) -> float:
    """Cubic calibration: a * x^3 + b * x^2 + c * x + d."""
    return a * x**3 + b * x**2 + c * x + d


def multilateration(
    raw_data: Dict[int, float],
    anchor_positions: Dict[int, Tuple[float, float, float]] = {},
    z_sign: int = 0,
) -> Tuple[float, float, float] | None:
    """
    Multilateration algorithm using localization library.

    :param raw_data: Dictionary mapping anchor IDs to distances (meters)
    :param anchor_positions: Dictionary mapping anchor IDs to (x, y, z) positions
    :param z_sign: z sign, used if there are two solutions (0 = auto, +1 = positive, -1 = negative)
    :return: (x, y, z) position tuple
    """
    P = lx.Project(mode="3D", solver="LSE")
    non_zero_data = {}

    for anchor_id in raw_data.keys():
        if raw_data[anchor_id] is not None:
            non_zero_data[anchor_id] = raw_data[anchor_id]
    logger.debug(f"non_zero_data {non_zero_data}")
    if len(non_zero_data) < 3:
        raise ValueError(f"Not enough data for trilateration. Non-zero: {non_zero_data}, Input: {raw_data}")
    logger.debug(f"anchor_positions {anchor_positions}")
    if len(anchor_positions) < 3:
        raise ValueError(f"Not enough anchor positions for trilateration: {anchor_positions}")

    matched_ids = list(set(non_zero_data.keys()) & set(anchor_positions.keys()))
    if len(matched_ids) < 3:
        raise ValueError(f"Not enough common anchors for trilateration. Matched: {matched_ids}")

    for anchor_id in matched_ids:
        P.add_anchor(anchor_id, anchor_positions[anchor_id])

    t, _ = P.add_target()
    for anchor_id in matched_ids:
        t.add_measure(anchor_id, non_zero_data[anchor_id])

    P.solve()

    if z_sign != 0:
        t.loc.z = abs(t.loc.z) * z_sign

    return t.loc.x, t.loc.y, t.loc.z


class LocalizationEngine:
    """Main localization engine that processes distance measurements and calculates position."""

    __attributes__ = [
        "anchor_positions",
        "origin_coordinates",
        "ellipsoid",
        "calibration_type",
        "calibration_params",
        "z_sign",
        "min_range",
        "max_range",
    ]

    def __init__(self, config: LocalizationConfig):
        """
        Initialize localization engine from configuration.

        :param config: Localization configuration
        """
        self.ellipsoid: pm.Ellipsoid = pm.Ellipsoid.from_name(config.ellipsoid)
        self.anchor_positions: Dict[int, Tuple[float, float, float]] = config.anchor_positions
        self.origin_coordinates: Optional[Tuple[float, float, float]] = config.origin_coordinates
        self.calibration_type: str = config.calibration_type
        self.calibration_params: List[float] = config.calibration_params
        self.z_sign: int = config.z_sign
        self.min_range: float = config.min_range
        self.max_range: float = config.max_range

    def calibrate(self, raw_distance: float) -> float:
        """Apply calibration to raw distance measurement."""
        if self.calibration_type == "linear":
            if len(self.calibration_params) != 2:
                raise ValueError("Linear calibration requires 2 parameters [k, b]")
            return calibrated_linear(self.calibration_params[0], self.calibration_params[1], raw_distance)
        if self.calibration_type == "quadratic":
            if len(self.calibration_params) != 3:
                raise ValueError("Quadratic calibration requires 3 parameters [a, b, c]")
            return calibrated_quadratic(
                self.calibration_params[0],
                self.calibration_params[1],
                self.calibration_params[2],
                raw_distance,
            )
        if self.calibration_type == "cubic":
            if len(self.calibration_params) != 4:
                raise ValueError("Cubic calibration requires 4 parameters [a, b, c, d]")
            return calibrated_cubic(
                self.calibration_params[0],
                self.calibration_params[1],
                self.calibration_params[2],
                self.calibration_params[3],
                raw_distance,
            )
        return raw_distance

    def validate_range(self, distance: float) -> bool:
        """Check if distance is within valid range."""
        return self.min_range <= distance <= self.max_range

    def calculate_position(self, distances: Dict[int, float]) -> Optional[Tuple[float, float, float]]:
        """
        Calculate position from distance measurements.

        :param distances: Dictionary mapping anchor IDs to calibrated distances (meters)
        :return: (x, y, z) position tuple or None if calculation fails
        """
        logger.debug(f"distances {distances}")
        # Filter out None values and validate ranges
        valid_distances = {}
        for anchor_id, distance in distances.items():
            if distance is not None and self.validate_range(distance):
                valid_distances[anchor_id] = distance
        logger.debug(f"valid_distances {valid_distances}")
        if len(valid_distances) < 3:
            return None

        try:
            return multilateration(valid_distances, self.anchor_positions, self.z_sign)
        except ValueError as e:
            logger.error("Error in multilateration: %s", e)
            return None

    def _convert_to_gps(self, position: Tuple[float, float, float]) -> Tuple[float, float, float]:
        """Convert position to GPS coordinates."""
        if self.origin_coordinates is None:
            raise ValueError("Origin coordinates are not set")
        lat, lon, alt = pm.enu2geodetic(
            position[0],
            position[1],
            position[2],
            self.origin_coordinates[0],
            self.origin_coordinates[1],
            self.origin_coordinates[2],
            ell=self.ellipsoid,
        )
        return lat, lon, alt



"""
Unit tests for core localization module.
"""

import math

import pytest

from multilat_solver.core import (
    LocalizationConfig,
    LocalizationEngine,
    calibrated_cubic,
    calibrated_linear,
    calibrated_quadratic,
    multilateration,
)


class TestCalibration:
    """Test calibration functions."""

    def test_linear_calibration(self):
        """Test linear calibration."""
        assert calibrated_linear(2.0, 1.0, 5.0) == 11.0  # 2*5 + 1
        assert calibrated_linear(1.0, 0.0, 10.0) == 10.0  # Identity

    def test_quadratic_calibration(self):
        """Test quadratic calibration."""
        # x^2 + 2x + 1 at x=2 should be 9
        assert calibrated_quadratic(1.0, 2.0, 1.0, 2.0) == 9.0

    def test_cubic_calibration(self):
        """Test cubic calibration."""
        # x^3 at x=2 should be 8
        assert calibrated_cubic(1.0, 0.0, 0.0, 0.0, 2.0) == 8.0


class TestMultilateration:
    """Test multilateration function."""

    def test_multilateration_sufficient_anchors(self):
        """Test multilateration with sufficient anchors."""
        anchor_positions = {1: (0.0, 0.0, 0.0), 2: (3.0, 0.0, 0.0), 3: (0.0, 3.0, 0.0)}

        # Point at (1,1,0)
        distances = {1: math.sqrt(2), 2: math.sqrt(5), 3: math.sqrt(5)}

        x, y, z = multilateration(distances, anchor_positions, z_sign=0)

        assert abs(x - 1.0) < 0.1
        assert abs(y - 1.0) < 0.1
        assert abs(z - 0.0) < 0.1

    def test_multilateration_insufficient_anchors(self):
        """Test multilateration fails with insufficient anchors."""
        anchor_positions = {1: (0.0, 0.0, 0.0), 2: (3.0, 0.0, 0.0)}

        distances = {1: 1.0, 2: 2.0}

        with pytest.raises(ValueError, match="Not enough data for trilateration"):
            multilateration(distances, anchor_positions)

    def test_multilateration_insufficient_data(self):
        """Test multilateration fails with insufficient distance data."""
        anchor_positions = {1: (0.0, 0.0, 0.0), 2: (3.0, 0.0, 0.0), 3: (0.0, 3.0, 0.0)}

        distances = {1: 1.0, 2: 2.0}

        with pytest.raises(ValueError, match="Not enough data"):
            multilateration(distances, anchor_positions)

    def test_multilateration_no_common_anchors(self):
        """Test multilateration fails when no common anchors."""
        anchor_positions = {1: (0.0, 0.0, 0.0), 2: (3.0, 0.0, 0.0), 3: (0.0, 3.0, 0.0)}

        distances = {4: 1.0, 5: 2.0, 6: 3.0}

        with pytest.raises(ValueError, match="Not enough common anchors"):
            multilateration(distances, anchor_positions)


class TestLocalizationEngine:
    """Test LocalizationEngine class."""

    @pytest.fixture
    def engine(self):
        """Create a test localization engine."""
        anchor_positions = {1: (0.0, 0.0, 0.0), 2: (3.0, 0.0, 0.0), 3: (0.0, 3.0, 0.0)}
        return LocalizationEngine(
            LocalizationConfig(
                anchor_positions=anchor_positions,
                calibration_type="none",
                calibration_params=[],
                z_sign=0,
                min_range=0.0,
                max_range=100.0,
            )
        )

    def test_engine_initialization(self, engine):
        """Test engine initialization."""
        assert len(engine.anchor_positions) == 3
        assert engine.calibration_type == "none"

    def test_engine_insufficient_anchors(self):
        """Test engine fails with insufficient anchors."""
        anchor_positions = {1: (0.0, 0.0, 0.0), 2: (3.0, 0.0, 0.0)}

        with pytest.raises(ValueError, match="At least 3 anchor positions"):
            LocalizationEngine(LocalizationConfig(anchor_positions=anchor_positions))

    def test_calibrate_none(self, engine):
        """Test no calibration."""
        assert engine.calibrate(5.0) == 5.0

    def test_calibrate_linear(self):
        """Test linear calibration."""
        anchor_positions = {1: (0, 0, 0), 2: (3, 0, 0), 3: (0, 3, 0)}
        engine = LocalizationEngine(
            LocalizationConfig(
                anchor_positions=anchor_positions,
                calibration_type="linear",
                calibration_params=[2.0, 1.0],
            )
        )
        assert engine.calibrate(5.0) == 11.0  # 2*5 + 1

    def test_calibrate_quadratic(self):
        """Test quadratic calibration."""
        anchor_positions = {1: (0, 0, 0), 2: (3, 0, 0), 3: (0, 3, 0)}
        engine = LocalizationEngine(
            LocalizationConfig(
                anchor_positions=anchor_positions,
                calibration_type="quadratic",
                calibration_params=[1.0, 2.0, 1.0],
            )
        )
        assert engine.calibrate(2.0) == 9.0  # 2^2 + 2*2 + 1

    def test_validate_range(self, engine):
        """Test range validation."""
        assert engine.validate_range(50.0) is True
        assert engine.validate_range(0.0) is True
        assert engine.validate_range(100.0) is True
        assert engine.validate_range(-1.0) is False
        assert engine.validate_range(101.0) is False

    def test_calculate_position_valid(self, engine):
        """Test position calculation with valid data."""
        distances = {1: math.sqrt(2), 2: math.sqrt(5), 3: math.sqrt(5)}

        position = engine.calculate_position(distances)
        assert position is not None
        assert len(position) == 3
        x, y, z = position
        assert abs(x - 1.0) < 0.1
        assert abs(y - 1.0) < 0.1
        assert abs(z - 0.0) < 0.1

    def test_calculate_position_insufficient_data(self, engine):
        """Test position calculation with insufficient data."""
        distances = {1: 1.0, 2: 2.0}

        position = engine.calculate_position(distances)
        assert position is None

    def test_calculate_position_out_of_range(self, engine):
        """Test position calculation filters out-of-range distances."""
        distances = {1: math.sqrt(2), 2: math.sqrt(5), 3: 150.0}  # Out of range

        position = engine.calculate_position(distances)
        # Should still work with 2 valid distances if algorithm allows
        # But multilateration requires 3, so should return None
        assert position is None

    def test_calculate_position_none_values(self, engine):
        """Test position calculation handles None values."""
        distances = {1: math.sqrt(2), 2: None, 3: math.sqrt(5)}

        position = engine.calculate_position(distances)
        assert position is None  # Not enough valid distances



#!/usr/bin/env python3
"""
Universal MultiLat Localization Application.

Supports flexible input/output adapters for distance measurements and position output.
"""

import logging
import signal
import sys
import time
from typing import Dict, List, Tuple

import pymap3d as pm

from .config import MultiLatLocalizerConfig
from .core import LocalizationEngine
from .input_adapters import InputAdapter
from .output_adapters import ConsoleOutputAdapter, MultiOutputAdapter, OutputAdapter

logger = logging.getLogger(__name__)


class MultiLatLocalizerApp:
    """Main application class."""

    def __init__(self, config: MultiLatLocalizerConfig):
        """
        Initialize application with configuration.

        :param config: MultiLatLocalizerConfig instance
        """
        self.config: MultiLatLocalizerConfig = config
        self.engine: LocalizationEngine | None = None
        self.input_adapter: InputAdapter | None = None
        self.output_adapter: OutputAdapter | None = None
        self.running: bool = False
        self.last_position: Tuple[float, float, float] | None = None
        self.position_history: List[Tuple[float, float, float]] = []
        self.distances: Dict[int, Dict[str, float]] = {}
        self.last_timestamp: float = 0.0

    def _exit_signal_handler(self) -> None:
        """Handle shutdown signals."""
        logger.info("Shutting down...")
        self.stop()
        sys.exit(0)

    def _register_signal_handlers(self) -> None:
        """Register signal handlers."""
        signal.signal(signal.SIGINT, self._exit_signal_handler)
        signal.signal(signal.SIGTERM, self._exit_signal_handler)

    def _create_msg_adapters(self) -> None:
        """Create message adapters."""
        # Create input adapter
        self.input_adapter = self.config.input.adapter
        self.input_adapter.start(self._on_distance_update)
        logger.info("Input adapter started: %s", type(self.input_adapter).__name__)

        # Create output adapter
        self.output_adapter = self._create_output_adapter()
        logger.info("Output adapter initialized: %s", type(self.output_adapter).__name__)

    def _create_localization_engine(self) -> LocalizationEngine:
        """Create localization engine from config."""
        loc_config = self.config.localization

        return LocalizationEngine(loc_config)

    def _create_output_adapter(self) -> MultiOutputAdapter:
        """Create output adapter(s) from config."""
        output_config = self.config.output

        # Handle multiple outputs
        adapter_types = output_config.types
        if isinstance(adapter_types, str):
            adapter_types = [adapter_types]

        adapters = []

        for adapter_type in adapter_types:
            if adapter_type not in output_config.adapters:
                raise ValueError(f"Unknown adapter type: {adapter_type}")
            adapter = output_config.adapters[adapter_type]
            adapters.append(adapter)
        if not adapters:
            adapters.append(ConsoleOutputAdapter())

        if len(adapters) == 1:
            return adapters[0]
        return MultiOutputAdapter(adapters)

    def _clear_outdated_records(self, timestamp: float) -> Dict[int, Dict[str, float]]:
        """Clear outdated records from the distances dictionary."""
        for anchor_id, record in list(self.distances.items()):
            if record["timestamp"] < timestamp - self.config.localization.data_expiration_timeout_ms:
                del self.distances[anchor_id]
        return self.distances

    def _on_distance_update(self, distances: Dict[int, float]) -> None:
        """Handle distance update."""
        if not self.engine:
            return
        logger.debug(f"Distances: {distances}")
        # Apply calibration
        last_timestamp = 0
        anchor_ids = set(list(self.distances.keys()) + list(distances.keys()))
        if len(anchor_ids) == 0:
            return
        for anchor_id in anchor_ids:
            if anchor_id in distances:
                if distances[anchor_id]["distance"] <= self.config.localization.max_range:
                    self.distances[anchor_id] = distances[anchor_id]
                    continue
            if anchor_id in self.distances:
                self.distances.pop(anchor_id)

        calibrated_distances = {
            anchor_id: self.engine.calibrate(record["distance"])
            for anchor_id, record in self.distances.items()
        }
        logger.debug(f"Calibrated distances: {calibrated_distances}")
        logger.info("Max range: %s", self.config.localization.max_range)
        # Calculate position
        position = self.engine.calculate_position(calibrated_distances)
        print(f"Number of distances: {len(calibrated_distances)}")
        if position:
            self.position_history.append(position)
            self._send_position(position, last_timestamp)
            self._send_gps(position, last_timestamp, len(calibrated_distances))
            self.last_position = position

            if len(self.position_history) > 1000:  # Keep last 1000 positions
                self.position_history.pop(0)
        else:
            logger.warning("Could not calculate position from distances: %s", calibrated_distances)

    def _send_position(self, position: Tuple[float, float, float], timestamp: float) -> None:
        """Send position to output adapter."""
        if self.output_adapter:
            self.output_adapter.send_position(position, timestamp)

    def _send_gps(self, position: Tuple[float, float, float], timestamp: float, n_satellites: int) -> None:
        """Send GPS to output adapter."""
        if self.output_adapter:
            assert isinstance(self.engine, LocalizationEngine)
            # Convert ENU to GPS if origin is set
            if self.engine.origin_coordinates:
                lat, lon, alt = pm.enu2geodetic(
                    position[0],
                    position[1],
                    position[2],
                    self.engine.origin_coordinates[0],
                    self.engine.origin_coordinates[1],
                    self.engine.origin_coordinates[2],
                    ell=self.engine.ellipsoid,
                )
                self.output_adapter.send_gps((lat, lon, alt), timestamp, n_satellites)
            else:
                # If no origin, send position as-is (assuming it's already GPS)
                self.output_adapter.send_gps(position, timestamp, n_satellites)

    def start(self) -> None:
        """Start the localization system."""
        if not logging.getLogger().handlers:
            logging.basicConfig(
                level=logging.INFO,
                format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        logger.info("Initializing MultiLat Localization System...")

        # Create localization engine
        self.engine = self._create_localization_engine()
        logger.info(
            "Localization engine initialized with %s anchors",
            len(self.engine.anchor_positions),
        )

        # Register message publishers
        self._create_msg_adapters()

        self.running = True
        logger.info("MultiLat Localization System started. Press Ctrl+C to stop.")

    def stop(self) -> None:
        """Stop the localization system."""
        self.running = False

        if self.input_adapter:
            self.input_adapter.stop()
            logger.info("Input adapter stopped")

        if self.output_adapter:
            self.output_adapter.close()
            logger.info("Output adapter closed")

        logger.info("MultiLat Localization System stopped")

    def run(self) -> None:
        """Run the main loop."""
        self.start()

        try:
            while self.running:
                time.sleep(0.1)
        except KeyboardInterrupt:
            pass
        finally:
            self.stop()



"""
MultiLat Localization System - Flexible input/output architecture.
"""

from .app import MultiLatLocalizerApp
from .common_types import MessageType
from .config import (
    CalibrationType,
    ConfigBuilder,
    InputConfig,
    LocalizationConfig,
    MultiLatLocalizerConfig,
    SumOutputConfig,
)
from .core import LocalizationEngine, multilateration
from .input_adapters import INPUT_ADAPTERS  # noqa: F401
from .input_adapters import register_input_adapter  # noqa: F401
from .input_adapters import (
    FileInputAdapter,
    InputAdapter,
    MQTTInputAdapter,
    SerialInputAdapter,
    UDPInputAdapter,
)
from .output_adapters import OUTPUT_ADAPTERS  # noqa: F401
from .output_adapters import register_output_adapter  # noqa: F401
from .output_adapters import (
    ConsoleOutputAdapter,
    FileOutputAdapter,
    MavlinkOutputAdapter,
    MavrosOutputAdapter,
    MultiOutputAdapter,
    OutputAdapter,
    UDPOutputAdapter,
)

__version__ = "0.1.0"

__all__ = [
    # Core
    "LocalizationEngine",
    "multilateration",
    # App
    "MultiLatLocalizerApp",
    # Config
    "MultiLatLocalizerConfig",
    "ConfigBuilder",
    "LocalizationConfig",
    "InputConfig",
    "SumOutputConfig",
    "CalibrationType",
    "MessageType",
    # Input Adapters
    "InputAdapter",
    "SerialInputAdapter",
    "MQTTInputAdapter",
    "UDPInputAdapter",
    "FileInputAdapter",
    # Output Adapters
    "OutputAdapter",
    "MavlinkOutputAdapter",
    "MavrosOutputAdapter",
    "UDPOutputAdapter",
    "FileOutputAdapter",
    "ConsoleOutputAdapter",
    "MultiOutputAdapter",
]



"""Configuration API for MultiLat Localizer.

Provides programmatic configuration instead of config.toml files.
"""

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Union

from multilat_solver.common_types import CalibrationType, MessageType
from multilat_solver.input_adapters import (
    INPUT_ADAPTERS,
    FileInputAdapter,
    InputAdapter,
    SerialInputAdapter,
    UDPInputAdapter,
    register_input_adapter,
)
from multilat_solver.output_adapters import OutputAdapter, register_output_adapter

logger = logging.getLogger(__name__)


@dataclass
class LocalizationConfig:
    """Configuration for localization engine.

    @attributes:
        anchor_positions:
            Dictionary of anchor positions.
        origin_coordinates:
            Origin coordinates of the localizer in the format (latitude, longitude, altitude).
        ellipsoid:
            Ellipsoid of the localizer. Default is "wgs84".
        calibration_type:
            Calibration type. Default is "none". "none" for no calibration,\
                "linear" for linear calibration,\
                "quadratic" for quadratic calibration,\
                "cubic" for cubic calibration.
        calibration_params:
            Calibration parameters in the format [k, b] for linear calibration,\
                [a, b, c] for quadratic calibration,\
                [a, b, c, d] for cubic calibration.
        z_sign:
            Z sign. Default is 0.
        min_range:
            Minimum range in meters. Default is 0.0.
        max_range:
            Maximum range in meters. Default is 100.0.
    """

    anchor_positions: Dict[int, Tuple[float, float, float]]
    origin_coordinates: Optional[Tuple[float, float, float]] = (0.0, 0.0, 0.0)
    ellipsoid: str = "wgs84"  # "wgs84" or "ecef"
    calibration_type: CalibrationType = CalibrationType.NONE
    calibration_params: List[float] = field(default_factory=lambda: [1.0, 0.0])
    z_sign: int = 0
    min_range: float = 0.0
    max_range: float = 100.0
    data_expiration_timeout_ms: float = 1000.0

    def __post_init__(self):
        """Validate configuration after initialization."""
        if len(self.anchor_positions) < 3:
            raise ValueError("At least 3 anchor positions are required")

        # Validate calibration params
        if self.calibration_type == CalibrationType.LINEAR and len(self.calibration_params) != 2:
            raise ValueError("Linear calibration requires 2 parameters [k, b]")

        elif self.calibration_type == CalibrationType.QUADRATIC and len(self.calibration_params) != 3:
            raise ValueError("Quadratic calibration requires 3 parameters [a, b, c]")

        elif self.calibration_type == CalibrationType.CUBIC and len(self.calibration_params) != 4:
            raise ValueError("Cubic calibration requires 4 parameters [a, b, c, d]")


@dataclass
class InputConfig:
    """Input adapter configuration wrapper."""

    type_name: str  # Adapter type name (e.g., "serial", "mqtt", "udp", "file", or custom)
    adapter: Any = None


@dataclass
class SumOutputConfig:
    """Summary output adapter configuration wrapper."""

    types: List[str] = field(default_factory=list)  # Single type or list for multiple adapters
    adapters: Dict[str, Any] = field(default_factory=dict)
    # Adapter-specific configs (only one will be used based on type)
    configs: Dict[str, Any] = field(default_factory=dict)


@dataclass
class MultiLatLocalizerConfig:
    """Complete configuration for MultiLat Localizer."""

    localization: LocalizationConfig
    input: InputConfig
    output: Any

    @classmethod
    def from_dict(cls, config_dict: dict) -> "MultiLatLocalizerConfig":
        """Create configuration from dictionary (for backward compatibility with TOML)."""
        # Parse localization config
        loc_dict = config_dict.get("localization", {})
        anchor_positions = {
            int(k): tuple(v) if isinstance(v, list) else v
            for k, v in loc_dict.get("anchor_positions", {}).items()
        }

        origin_coords = loc_dict.get("origin_coordinates")
        if origin_coords and isinstance(origin_coords, dict):
            origin_coords = (
                origin_coords.get("lat", 0.0),
                origin_coords.get("lon", 0.0),
                origin_coords.get("alt", 0.0),
            )
        else:
            origin_coords = (0.0, 0.0, 0.0)

        localization = LocalizationConfig(
            anchor_positions=anchor_positions,
            origin_coordinates=origin_coords,
            ellipsoid=loc_dict.get("ellipsoid", "wgs84"),
            calibration_type=CalibrationType(loc_dict.get("calibration_type", "none")),
            calibration_params=loc_dict.get("calibration_params", [1.0, 0.0]),
            z_sign=loc_dict.get("z_sign", 0),
            min_range=loc_dict.get("min_range", 0.0),
            max_range=loc_dict.get("max_range", 100.0),
        )

        # Parse input config
        input_dict = config_dict.get("input", {})
        input_type = input_dict.get("str", "serial").lower()

        # Check if adapter type is registered
        if input_type not in INPUT_ADAPTERS:
            raise ValueError(
                f"Unknown input adapter type: {input_type}. "
                f"Available types: {list(INPUT_ADAPTERS.keys())}"
                f"Register adapter using register_input_adapter"
                f"Example: register_input_adapter('my', MyInputAdapter, MyInputConfig(my_param=1))"
            )

        # Map known adapter types to their config classes for backward compatibility
        config_class_map = {
            "serial": SerialInputAdapter,
            "udp": UDPInputAdapter,
            "file": FileInputAdapter,
        }

        # Use predefined config class if available, otherwise use dict directly
        if input_type in config_class_map:
            config_class = config_class_map[input_type]
            if input_type == "serial":
                adapter_obj = config_class(
                    port=input_dict.get("port", "/dev/ttyUSB0"),
                    baud=input_dict.get("baud", 460800),
                    timeout=input_dict.get("timeout", 1.0),
                )
            elif input_type == "udp":
                adapter_obj = config_class(
                    host=input_dict.get("host", "0.0.0.0"),
                    port=input_dict.get("port", 5005),
                )
            elif input_type == "file":
                adapter_obj = config_class(
                    filepath=input_dict.get("filepath", "distances.json"),
                    poll_interval=input_dict.get("poll_interval", 1.0),
                )
        else:
            # For custom adapters, use dict directly (will be converted in app.py)
            adapter_obj = {k: v for k, v in input_dict.items() if k != "str"}

        input_config = InputConfig(type_name=input_type, adapter=adapter_obj)

        # Parse output config
        output_dict = config_dict.get("output", {})
        output_str = output_dict.get("str", "console").lower()

        # Handle multiple output types
        if isinstance(output_str, str) and "," in output_str:
            output_strs = [t.strip() for t in output_str.split(",")]
        elif isinstance(output_str, list):
            output_strs = output_str
        else:
            output_strs = [output_str]

        output_config = SumOutputConfig(
            types=output_strs,
            adapters=output_dict.get("adapters", {}),
            configs=output_dict.get("configs", {}),
        )

        return cls(
            localization=localization,
            input=input_config,
            output=output_config,
        )


class ConfigBuilder:
    """Builder class for creating MultiLat Localizer configurations programmatically."""

    def __init__(self):
        """Initialize ConfigBuilder."""
        self.__localization: LocalizationConfig | None = None
        self.__input: InputConfig | None = None
        self.__output: SumOutputConfig | None = None
        self._message_type: MessageType = MessageType.BOTH

    def with_localization(
        self,
        localization: LocalizationConfig | None = None,
        anchor_positions: Dict[int, Tuple[float, float, float]] = {},
        origin_coordinates: Optional[Tuple[float, float, float]] = (0.0, 0.0, 0.0),
        ellipsoid: str = "wgs84",
        calibration_type: Union[CalibrationType, str] = CalibrationType.NONE,
        calibration_params: Optional[List[float]] = [],
        z_sign: int = 0,
        min_range: float = 0.0,
        max_range: float = 100.0,
    ) -> "ConfigBuilder":
        """Configure localization engine."""
        if localization is not None:
            self.__localization = localization
            return self

        if isinstance(calibration_type, str):
            calibration_type = CalibrationType(calibration_type)
        logger.debug(f"anchor_positions {anchor_positions}")
        self.__localization = LocalizationConfig(
            anchor_positions=anchor_positions,
            origin_coordinates=origin_coordinates,
            ellipsoid=ellipsoid,
            calibration_type=calibration_type,
            calibration_params=calibration_params or [1.0, 0.0],
            z_sign=z_sign,
            min_range=min_range,
            max_range=max_range,
        )
        return self

    def with_serial_input(
        self,
        port: str = "/dev/ttyUSB0",
        baud: int = 460800,
        timeout: float = 1.0,
    ) -> "ConfigBuilder":
        """Configure serial input adapter."""
        self.__input = InputConfig(
            type_name="serial",
            adapter=SerialInputAdapter(port=port, baud=baud, timeout=timeout),
        )
        return self

    def with_udp_input(
        self,
        host: str = "0.0.0.0",
        port: int = 5005,
    ) -> "ConfigBuilder":
        """Configure UDP input adapter."""
        self.__input = InputConfig(type_name="udp", adapter=UDPInputAdapter(host=host, port=port))
        return self

    def with_file_input(
        self,
        filepath: str = "distances.json",
        poll_interval: float = 1.0,
    ) -> "ConfigBuilder":
        """Configure file input adapter."""
        self.__input = InputConfig(
            type_name="file",
            adapter=FileInputAdapter(filepath=filepath, poll_interval=poll_interval),
        )
        return self

    def with_input(self, input_str: str, adapter: Any) -> "ConfigBuilder":
        """Configure input adapter."""
        if self.__input is None:
            self.__input = InputConfig(type_name=input_str)
        if not issubclass(type(adapter), InputAdapter):
            raise ValueError("Adapter must be a subclass of InputAdapter")
        self.__input.adapter = adapter
        return self

    def with_output(self, output_str: str, adapter: Any) -> "ConfigBuilder":
        """Configure output adapter."""
        if self.__output is None:
            self.__output = SumOutputConfig()
        if not issubclass(type(adapter), OutputAdapter):
            raise ValueError("Adapter must be a subclass of BaseOutputConfig")
        self.__output.types.append(output_str)
        self.__output.adapters[output_str] = adapter
        return self

    def register_output(self, output_str: str, adapter: Any) -> "ConfigBuilder":
        """Register output adapter."""
        if adapter is None:
            raise ValueError("Adapter cannot be None")
        if not issubclass(type(adapter), OutputAdapter):
            raise ValueError("Adapter must be a subclass of OutputAdapter")
        register_output_adapter(output_str, adapter)
        if self.__output is None:
            self.__output = SumOutputConfig()
        self.__output.adapters[output_str] = adapter
        return self

    def register_input(self, adapter_type: str, adapter: Any, config: Any = None) -> "ConfigBuilder":
        """Register input adapter."""
        register_input_adapter(adapter_type, adapter)
        if self.__input is None:
            self.__input = InputConfig(type_name=adapter_type, config=config.__dict__)
        return self

    def build(self) -> MultiLatLocalizerConfig:
        """Build the configuration."""
        if self.__localization is None:
            raise ValueError("Localization configuration is required")
        if self.__input is None:
            raise ValueError("Input configuration is required")
        if self.__output is None:
            raise ValueError("Output configuration is required")

        assert issubclass(type(self.__input.adapter), InputAdapter)

        return MultiLatLocalizerConfig(
            localization=self.__localization,
            input=self.__input,
            output=self.__output,
        )
